import argparse
import dataclasses
import datetime
import io
import shlex
import typing

import colorama
import prompt_toolkit
from prompt_toolkit.formatted_text import ANSI
from prompt_toolkit.history import InMemoryHistory

import reviser
from reviser import commands
from reviser import definitions
from reviser import templating
from reviser.interactivity import completions


@dataclasses.dataclass(frozen=True)
class ExecutionResult:
    """Data structure for shell command execution responses."""

    status: str
    message: str
    info: typing.Optional[dict] = None
    data: typing.Optional[dict] = None


@dataclasses.dataclass(frozen=True)
class Execution:
    """Data structure for a shell command execution."""

    action: str
    args: dict
    shell: "Shell"
    result: typing.Optional["ExecutionResult"] = None

    executed_at: datetime.datetime = dataclasses.field(
        init=False,
        default_factory=lambda: datetime.datetime.utcnow(),
    )

    @property
    def timestamp(self) -> str:
        """A string representation of the created at datetime."""
        return self.executed_at.strftime("%Y-%m-%dT%H:%M:%S.000Z")

    def finalize(
        self,
        status: str,
        message: str = None,
        info: dict = None,
        data: dict = None,
        echo: bool = False,
    ) -> "Execution":
        """Copy this execution and populate it with the result."""
        return dataclasses.replace(
            self,
            result=ExecutionResult(
                status=(status or "???").upper(),
                message=message,
                info=info,
                data=data,
            ),
        ).echo_if(echo)

    def echo(self) -> "Execution":
        """Echoes the result if set for display to the console."""
        if self.result:
            templating.printer(
                "interactivity/shells/execution_result.jinja2",
                result=self.result,
            )
        return self

    def echo_if(self, condition) -> "Execution":
        """Echoes the result if set for display to the console."""
        if bool(condition):
            return self.echo()
        return self


class Shell:
    """
    Interactive lambda deployer interactivity class responsible for managing
    the state and interactivity for the interface.
    """

    def __init__(
        self,
        context: "definitions.Context",
        selection: "definitions.Selection" = None,
    ):
        """Creates a new shell for queued and/or interactive execution."""
        self.command_history: typing.List[str] = []
        self.execution_history: typing.List["Execution"] = []
        self._prompt_session: typing.Optional[prompt_toolkit.PromptSession] = None
        self.context = context
        self.selection = selection or definitions.Selection()
        self.shutdown = False
        #: Stores uncaught exceptions for reference when a command fails.
        self.error: typing.Optional[Exception] = None
        self._shell_completer = completions.ShellCompleter(self)
        self.command_queue: typing.List[str] = []
        #: Explicitly specifies if the shell should be running in
        #: interactive mode where it prompts users for input on the
        #: terminal.
        self._is_interactive: typing.Optional[bool] = None

    @property
    def is_interactive(self) -> typing.Optional[bool]:
        """
        Whether or not the shell is running as an interactive
        command loop the prompts users for input.
        """
        return self._is_interactive

    @is_interactive.setter
    def is_interactive(self, value: typing.Optional[bool]):
        """Sets the is_interactive property."""
        if value and self._prompt_session is None:
            if not self._create_prompt_session():
                raise RuntimeError("Terminal connectivity not supported.")

        self._is_interactive = value

    @property
    def is_active(self) -> bool:
        """
        Whether or not the shell should continue running or not as
        determined by whether or not it is running interactively or
        there are queued commands to still process.
        """
        return not self.shutdown and (bool(self.command_queue) or self.is_interactive)

    def _create_prompt_session(self) -> bool:
        """
        Attempts to create a session for the interactive command execution
        prompts and returns whether or not that session was successfully
        created. It will fail when executed in non-interactive environments
        that do no support user input.
        """
        history = InMemoryHistory()
        for line in self.command_history:
            history.append_string(line)

        try:
            self._prompt_session = prompt_toolkit.PromptSession(history=history)
        except io.UnsupportedOperation:
            self._is_interactive = False
            return False

        return True

    def _setup(self):
        """Initialization before entering the command loop."""
        templating.printer(
            "interactivity/shells/splash.jinja2",
            version=reviser.__version__,
        )
        print("\n\n")
        colorama.init()

        # If no queued commands have been specified at the start of the
        # loop, consider the loop to be running in interactive mode.
        self.is_interactive = not bool(self.command_queue)

    def run(self) -> typing.NoReturn:
        """Launches the command execution loop."""
        self._setup()

        try:
            while not self.shutdown and self.is_active:
                if line := self._get_next_command():
                    execute(self, line)
        except KeyboardInterrupt:
            print("\n[INTERRUPTED]: Shutting down terminal.")
        except Exception as error:
            self.error = error
            templating.print_error(
                error=error,
                message="An unexpected command error occurred.",
            )

        self._teardown()

    def _teardown(self):
        """Teardown after exiting the command loop."""
        print("\n\n")
        colorama.deinit()
        self.shutdown = True
        self.command_queue = []
        self.is_interactive = False
        self._prompt_session = None

    def _get_next_command(self) -> str:
        """
        Prompts user for input and returns that for command execution
        unless there is a queued command, in which case that is returned
        without a prompt instead.
        """
        context = self.context
        templating.printer(
            "interactivity/shells/prompt.jinja2",
            profile=context.connection.session.profile_name or "default",
            user_slug=context.connection.user_slug,
            selected=context.get_selected_targets(self.selection),
        )
        prompt = f"{colorama.Fore.GREEN}>{colorama.Style.RESET_ALL} "

        if self.command_queue:
            line = self.command_queue.pop(0)
            print(f"{prompt}{line}")
        else:
            # noinspection PyTypeChecker
            line = self._prompt_session.prompt(
                message=ANSI(prompt),
                completer=self._shell_completer,
                complete_while_typing=True,
            )

        return line.strip()


def execute(shell: "Shell", line: str) -> typing.NoReturn:
    """
    Executes the specified input command within the given shell
    environment.
    """
    if len(line.strip()) < 1:
        return

    try:
        raw_args = shlex.split(line)
        action = raw_args[0]
        raw_args = raw_args[1:]
    except Exception as error:
        shell.error = error
        templating.print_error(
            error=error,
            message=f'An error occurred parsing "{line}".',
        )
        return

    action_module = commands.get_module(action)

    print("\n")
    if action_module is None:
        templating.print_error(f'Unknown command "{action}".')
        return

    try:
        try:
            parser = argparse.ArgumentParser(prog=action)
            action_module.populate_subparser(parser)
            args = vars(parser.parse_args(raw_args))
        except SystemExit:
            return

        execution = action_module.run(Execution(action, args, shell))
        shell.execution_history.append(execution)
        shell.command_history.append(line)
    except Exception as error:
        shell.error = error
        templating.print_error(
            error=error,
            message="An unexpected command error occurred.",
        )
    finally:
        print("\n")



import dataclasses
import glob
import pathlib
import shutil
import typing

from reviser.definitions import abstracts
from reviser.definitions import configurations
from reviser.definitions import enumerations


def _get_paths(patterns: typing.Set[pathlib.Path]) -> typing.Set[pathlib.Path]:
    """
    Converts a set of file/directory patterns into a list of matching.
    files.
    """
    raw = [
        pathlib.Path(item)
        for pattern in patterns
        for item in glob.iglob(str(pattern), recursive=True)
    ]
    files = [p for p in raw if not p.is_dir()]
    output = files + [
        child
        for p in raw
        for item in glob.iglob(f"{p}/**/*", recursive=True)
        if p.is_dir() and not (child := pathlib.Path(item)).is_dir()
    ]
    return set(output)


@dataclasses.dataclass(frozen=True)
class CopyPath:
    """Data structure for a file path copy from source to target."""

    source: pathlib.Path
    destination: pathlib.Path

    def copy(self) -> bool:
        """Copies the file from source to destination."""
        if not self.source.exists():
            return False
        self.destination.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy2(
            src=str(self.source.absolute()),
            dst=str(self.destination.absolute()),
        )


@dataclasses.dataclass(frozen=True)
class Bundle(abstracts.Specification):
    """
    Defines the function or layer bundle that will be generated, i.e. built,
    and deployed as a zipped artifact to the lambda function or layer target.
    """

    target: "configurations.Target"

    @property
    def handler(self) -> typing.Optional[str]:
        """
        Function handler in the format `filename.function_name`. If not
        specified explicitly, it will default to the lambda function
        default value of `lambda_function.lambda_handler`.
        """
        if self.target.kind == enumerations.TargetType.LAYER:
            return None
        return self.get("handler", default="lambda_function.lambda_handler")

    @property
    def handler_filename(self) -> typing.Optional[str]:
        """Python filename associated with the specified handler."""
        if handler := self.handler:
            return "{}.py".format(handler.rsplit(".", 1)[0])

    @property
    def handler_function(self) -> typing.Optional[str]:
        """Python entrypoint function name for the specified handler."""
        if handler := self.handler:
            return handler.rsplit(".", 1)[-1]

    @property
    def omitted_packages(self) -> typing.List[str]:
        """
        A list of site-packages to skip with creating the bundle. This
        is especially useful when package dependencies are installed
        in both a function and layer definition and should be omitted from
        one such that they are not duplicated in the final lambda function
        environment.
        """
        omissions = self.get("omit_packages", default=self.get("omit_package"))
        if isinstance(omissions, str):
            return [omissions]

        return omissions or []

    @property
    def file_include_patterns(self) -> typing.Set[pathlib.Path]:
        """
        A list of file matching patterns that will be used to determine
        what to copy into the bundled artifact deployed to lambda for the
        given target. These should be defined relative the root directory
        defined by the location of the configuration file and stored in
        the directory value of this object.
        """
        includes = self.get("includes", default=self.get("include")) or []
        if isinstance(includes, str):
            includes = [includes]

        add_default_package = (
            not includes and self.target.kind == enumerations.TargetType.FUNCTION
        )
        if add_default_package:
            # If the user hasn't specified any includes, look for python
            # packages in the root directory and include those by default
            # for functions. By default, layers should include nothing.
            includes = [
                pathlib.Path(item).parent.absolute()
                for item in glob.iglob(
                    str(self.directory.joinpath("*/__init__.py")),
                    recursive=True,
                )
            ]

        if filename := self.handler_filename:
            # The handler file should always be included if set.
            includes.append(self.directory.joinpath(filename))

        return set([self.directory.joinpath(item).absolute() for item in includes])

    @property
    def file_exclude_patterns(self) -> typing.Set[pathlib.Path]:
        """
        A list of file matching patterns that will be used to determine
        what to skip copying into the bundled artifact deployed to lambda
        for the given target. These should be defined relative the root
        directory defined by the location of the configuration file and
        stored in the directory value of this object. Note that exclusions
        are applied to the inclusion results after so only need to specify
        things to remove from the inclusion list, e.g. a `tests` folder.
        """
        excludes = self.get("excludes", default=self.get("exclude")) or []
        if isinstance(excludes, str):
            excludes = [excludes]

        # Exclude Python and OS cache files by default.
        excludes += ["**/__pycache__", "**/*.pyc", "**/.DS_Store"]

        return set([self.directory.joinpath(item).absolute() for item in excludes])

    def get_include_paths(
        self,
        relative: bool = False,
    ) -> typing.Set[pathlib.Path]:
        """Generates a list of file paths matching the bundle includes."""
        items = _get_paths(self.file_include_patterns)
        if filename := self.handler_filename:
            items.add(self.directory.joinpath(filename))
        if relative:
            return set([item.relative_to(self.directory) for item in items])
        return items

    def get_exclude_paths(
        self,
        relative: bool = False,
    ) -> typing.Set[pathlib.Path]:
        """Generates a list of file paths matching the bundle excludes."""
        items = _get_paths(self.file_exclude_patterns)
        if relative:
            return set([item.relative_to(self.directory) for item in items])
        return items

    def get_paths(
        self,
        relative: bool = False,
    ) -> typing.Set[pathlib.Path]:
        """Generates a list of file paths matching the bundle excludes."""
        return self.get_include_paths(relative=relative) - self.get_exclude_paths(
            relative=relative
        )

    def get_copy_paths(self) -> typing.List["CopyPath"]:
        """
        A list of copy paths that map included (and not excluded) source
        paths to their respective destination path in the bundle directory.
        """
        return [
            CopyPath(
                source=self.directory.joinpath(p),
                destination=self.target.bundle_directory.joinpath(p),
            )
            for p in self.get_paths(relative=True)
        ]

    def get_site_package_paths(self) -> typing.Set[pathlib.Path]:
        """
        A list of paths in the target bundle site packages directory
        that should be copied into the bundle zip for deployment. Omitted
        packages will be excluded from this list if any are specified in
        the bundle.
        """
        directory = self.target.site_packages_directory
        exclusions = []
        for package in self.omitted_packages:
            exclusions += [
                directory.joinpath(package, "**", "*"),
                directory.joinpath(f"{package}.*"),
                directory.joinpath(f"{package}-*"),
                directory.joinpath(f"{package}-*", "**", "*"),
            ]
        omitted_paths = _get_paths(set(exclusions))

        paths = _get_paths({directory.joinpath("**", "*")})
        return paths - omitted_paths

    def serialize(self) -> dict:
        """Serializes the object for output representation."""
        return {
            "handler": self.handler,
            "handler_filename": self.handler_filename,
            "handler_function": self.handler_function,
            "omitted_packages": self.omitted_packages,
            "file_include_patterns": [str(p) for p in self.file_include_patterns],
            "file_exclude_patterns": [str(p) for p in self.file_exclude_patterns],
        }



import enum
import typing
import sys


RUNTIME_VERSION = f"{sys.version_info.major}.{sys.version_info.minor}"


class TargetType(enum.Enum):
    """Enumerations of target types."""

    FUNCTION = "function"
    LAYER = "layer"


class DependencyType(enum.Enum):
    """Enumeration of dependency types, i.e. package managers."""

    PIP = "pip"
    PIPPER = "pipper"
    POETRY = "poetry"


class DefaultFile(enum.Enum):
    """Enumeration of default files for each dependency type."""

    PIP = "requirements.txt"
    PIPPER = "pipper.json"
    POETRY = "pyproject.toml"

    @classmethod
    def from_dependency_type(
        cls,
        dependency_type: typing.Union["DependencyType", str],
    ) -> typing.Optional["DefaultFile"]:
        """Lookup default file from a dependency type."""
        key = getattr(dependency_type, "value", dependency_type)
        return {
            DependencyType.PIP.value: DefaultFile.PIP,
            DependencyType.PIPPER.value: DefaultFile.PIPPER,
            DependencyType.POETRY.value: DefaultFile.POETRY,
        }.get(key, None)



import dataclasses
import pathlib
import typing
import uuid

from reviser.definitions import aws


@dataclasses.dataclass(frozen=True)
class DataWrapper:
    """Base class for data loaded and stored in a dynamic dictionary."""

    #: Data that represents the dynamic representation that is being
    #: wrapped by this object.
    data: dict

    def get(self, *args: str, default: typing.Any = None) -> typing.Any:
        """Fetches the value from the data dictionary."""
        value = self.data or {}
        for k in args:
            if k not in value:
                return default
            value = value[k]
        return value

    def get_first(
        self,
        *args: typing.Iterable[str],
        default: typing.Any = None,
    ) -> typing.Any:
        """
        Fetches the value from the data dictionary in the key-grouped
        order, returning the default value if none of the key-group args
        exist in the data dictionary.
        """
        found_args = next((a for a in args if self.has(*a)), None)
        if found_args is None:
            return default
        return self.get(*found_args, default=default)

    def get_as_list(
        self,
        *args: str,
        default: typing.Any = None,
    ) -> typing.List[typing.Any]:
        """
        Fetches the value from the data dictionary as a list.
        None values will be returned as an empty list. Values that are
        not lists will be wrapped into single-length lists when returned.
        If the nested key does not exist, the default will be returned
        instead, but if the key is explicitly defined as `null` an empty
        list is returned instead fo the default.
        """
        if not self.has(*args):
            return default

        value = self.get(*args, default=default)
        if value is None:
            return []
        if isinstance(value, list):
            return value
        return [value]

    def get_first_as_list(
        self,
        *args: typing.Iterable[str],
        default: typing.Any = None,
    ) -> typing.Any:
        """
        Fetches the value from the data dictionary in the key-grouped
        order, returning the default value if none of the key-group args
        exist in the data dictionary. If a match is found it will be returned
        as a list in the same fashion as the get_as_list function.
        """
        found_args = next((a for a in args if self.has(*a)), None)
        if found_args is None:
            return default
        return self.get_as_list(*found_args, default=default)

    def has(self, *args: str) -> bool:
        """Determines whether or not the nested key exists."""
        value = self.data or {}
        for k in args:
            if k not in value:
                return False
            value = value[k]
        return True

    def update(self, *args, value: typing.Any = None, overwrite: bool = False):
        """
        Updates the data object if the key is not found to be set already.
        """
        container = self.data or {}
        for k in args[:-1]:
            if k not in container:
                container[k] = {}
            container = container[k]

        if overwrite or args[-1] not in container:
            container[args[-1]] = value


@dataclasses.dataclass(frozen=True)
class Specification(DataWrapper):
    """Base class for data loaded from the configuration."""

    #: Root directory in which the project being operated on resides.
    directory: pathlib.Path
    #: Specification data loaded from a configuration file that represents
    #: the data for this object.
    data: dict
    #: Connection object that represents the AWS session and account data
    #: associated with the context in which this configuration resides.
    connection: "aws.AwsConnection"

    def __post_init__(self):
        """
        Add an internal uuid to the object's data dictionary after creation
        if it has not already been set.
        """
        if not self.has("_uuid"):
            self.update("_uuid", value=str(uuid.uuid4()), overwrite=False)

    @property
    def uuid(self) -> str:
        """An internal unique identifier for the configuration object."""
        return self.get("_uuid")

    def serialize(self) -> dict:
        """Serializes the object for output representation."""
        return {}



from reviser.interactivity.completions import ShellCompleter  # noqa
from reviser.interactivity.shells import Execution  # noqa
from reviser.interactivity.shells import ExecutionResult  # noqa
from reviser.interactivity.shells import Shell  # noqa
from reviser.interactivity.runner import create_shell  # noqa
from reviser.interactivity.runner import run_shell  # noqa



"""
Shows the current status information for each of the selected lambda targets.
"""
import argparse
import textwrap
import typing

import yaml
from botocore.client import BaseClient

from reviser import definitions
from reviser import interactivity
from reviser import servicer


def get_completions(
    completer: "interactivity.ShellCompleter",
) -> typing.List[str]:
    """Shell auto-completes for this command."""
    return []


def populate_subparser(parser: argparse.ArgumentParser):
    """Populate parser for the status command."""
    parser.add_argument(
        "qualifier",
        nargs="?",
        help="""
        Specifies a version or alias to show status for.
        If not specified, $LATEST will be used for functions
        and the latest version will be dynamically determined
        for layers.
        """,
    )


def _get_layer_version_info(
    client: BaseClient,
    layer_reference: "definitions.LambdaLayerReference",
) -> dict:
    """Fetches layer information for display."""
    versions = servicer.get_layer_versions(
        client,
        layer_reference.unversioned_arn,
    )
    current = next((v for v in versions if v.arn == layer_reference.arn), None)
    latest = versions[-1]

    out = {
        "name": current.name,
        "version": current.version,
        "created": current.created.isoformat("T"),
        "runtimes": ", ".join(current.runtimes),
        "arn": layer_reference.arn,
    }

    if latest != current:
        out["status"] = f"Newer version {latest.version} exists."
    else:
        out["status"] = "Is latest version."

    return out


def _display_function_info(
    client: BaseClient,
    name: str,
    qualifier: str,
):
    """Displays the response lambda function information."""
    lambda_function = servicer.get_function_version(
        lambda_client=client,
        function_name=name,
        qualifier=qualifier,
    )
    data = {
        "modified": lambda_function.modified,
        "description": lambda_function.description,
        "arn": lambda_function.arn,
        "runtime": lambda_function.runtime,
        "role": lambda_function.role,
        "handler": lambda_function.handler,
        "size": lambda_function.size,
        "timeout": lambda_function.timeout,
        "memory": lambda_function.memory,
        "version": lambda_function.version,
        "environment": lambda_function.environment,
        "revision_id": lambda_function.revision_id,
        "layers": [
            {
                **_get_layer_version_info(client, item),
                "size": item.size,
            }
            for item in lambda_function.layers
        ],
        "status": lambda_function.status.to_dict(),
        "update_status": lambda_function.status.to_dict(),
    }
    suffix = qualifier or "$LATEST"
    print(f"\n--- {lambda_function.name}:{suffix} ---")
    print(textwrap.indent(yaml.safe_dump(data), prefix="  "))
    print("\n")


def _display_layer_info(
    client: BaseClient,
    name: str,
    qualifier: str,
):
    """Display layer version information."""
    try:
        version = int(qualifier)
    except (ValueError, TypeError):
        version = None

    if version is None:
        version = servicer.get_layer_versions(client, name)[-1].version

    layer = servicer.get_layer_version(
        lambda_client=client,
        layer_name=name,
        version=version,
    )

    print(f"\n--- {layer.name}:{version} ---")
    data = {
        "arn": layer.arn,
        "version": layer.version,
        "created": layer.created.isoformat(),
        "description": layer.description,
        "size": layer.size,
        "runtimes": ", ".join(layer.runtimes),
    }
    print(textwrap.indent(yaml.safe_dump(data), prefix="  "))


def run(ex: "interactivity.Execution") -> "interactivity.Execution":
    """Displays the current configuration of the lambda target(s)."""
    selected = ex.shell.context.get_selected_targets(ex.shell.selection)
    qualifier = ex.args.get("qualifier")

    items = [(t, n) for t in selected.function_targets for n in t.names]
    for target, name in items:
        _display_function_info(target.client("lambda"), name, qualifier)

    items = [(t, n) for t in selected.layer_targets for n in t.names]
    for target, name in items:
        _display_layer_info(target.client("lambda"), name, qualifier)

    return ex.finalize(
        status="SUCCESS",
        message="Status reports have been display.",
        echo=False,
    )



import dataclasses
import datetime
import typing

from reviser import utils


@dataclasses.dataclass(frozen=True)
class Status:
    """Data structure containing lambda status information."""

    state: typing.Optional[str]
    reason: typing.Optional[str]
    code: typing.Optional[str]

    def to_dict(self) -> dict:
        """
        Returns this data as a sparse dictionary containing only the non-None
        key/value pairs.
        """
        return {
            key: value
            for key in ["state", "reason", "code"]
            if (value := getattr(self, key)) is not None
        }


@dataclasses.dataclass(frozen=True)
class LambdaLayerReference:
    """Data structure containing reference information for a lambda layer."""

    response: dict

    @property
    def arn(self) -> typing.Optional[str]:
        return self.get("Arn")

    @property
    def unversioned_arn(self) -> typing.Optional[str]:
        """Returns the layer's unversioned ARN."""
        if not self.arn:
            return None
        parts = self.arn.split(":")
        return ":".join(parts[:7])

    @property
    def name(self) -> str:
        """The name of the layer extracted from its ARN."""
        return (self.arn or "arn:???:?").rsplit(":", 2)[-2]

    @property
    def version(self) -> str:
        """The version for the layer extracted from its ARN."""
        return (self.arn or "arn:???:?").rsplit(":", 1)[-1]

    @property
    def size(self) -> typing.Optional[str]:
        return utils.to_human_readable_size(self.response.get("CodeSize"))

    def get(self, key: str, default: typing.Any = None):
        """Fetches the value of the specified response key."""
        return (self.response or {}).get(key, default)


@dataclasses.dataclass(frozen=True)
class LambdaLayer:
    """Data structure for a `list_layer_version` AWS API response."""

    response: dict

    @property
    def arn(self) -> typing.Optional[str]:
        return self.get("LayerVersionArn")

    @property
    def unversioned_arn(self) -> typing.Optional[str]:
        if unversioned := self.get("LayerArn"):
            return unversioned
        if not self.arn:
            return None
        parts = self.arn.split(":")
        return ":".join(parts[:7])

    @property
    def name(self) -> str:
        """The name of the layer extracted from its ARN."""
        return (self.arn or "arn:???:?").rsplit(":", 2)[-2]

    @property
    def version(self) -> typing.Optional[int]:
        return self.get("Version")

    @property
    def description(self) -> typing.Optional[str]:
        return self.get("Description")

    @property
    def created(self) -> typing.Optional[datetime.datetime]:
        try:
            return datetime.datetime.fromisoformat(
                (self.response or {}).get("CreatedDate").rsplit("+", 1)[0]
            ).replace(microsecond=0)
        except Exception:
            return None

    @property
    def runtimes(self) -> typing.List[str]:
        return self.get("CompatibleRuntimes") or []

    @property
    def size(self) -> typing.Optional[str]:
        return utils.to_human_readable_size(self.response.get("CodeSize"))

    def get(self, key: str, default: typing.Any = None):
        """Fetches the value of the specified response key."""
        return (self.response or {}).get(key, default)


@dataclasses.dataclass(frozen=True)
class FunctionAlias:
    """Data structure for lambda function aliases."""

    response: dict

    @property
    def arn(self) -> typing.Optional[str]:
        return self.get("AliasArn")

    @property
    def name(self) -> typing.Optional[str]:
        return self.get("Name")

    @property
    def function_version(self) -> typing.Optional[str]:
        return self.get("FunctionVersion")

    @property
    def description(self) -> typing.Optional[str]:
        return self.get("Description")

    @property
    def revision_id(self) -> typing.Optional[str]:
        return self.get("RevisionId")

    def get(self, key: str, default: typing.Any = None):
        """Fetches the value of the specified response key."""
        return (self.response or {}).get(key, default)


@dataclasses.dataclass(frozen=True)
class LambdaFunction:
    """Data structure for a `get_function_configuration` AWS API response."""

    response: dict
    alias_responses: typing.Optional[typing.List[dict]] = None

    @property
    def name(self) -> typing.Optional[str]:
        return self.get("FunctionName")

    @property
    def aliases(self) -> typing.List["FunctionAlias"]:
        return [FunctionAlias(item) for item in (self.alias_responses or [])]

    @property
    def modified(self) -> typing.Optional[str]:
        return self.get("LastModified")

    @property
    def description(self) -> typing.Optional[str]:
        return self.get("Description")

    @property
    def arn(self) -> typing.Optional[str]:
        return self.get("FunctionArn")

    @property
    def runtime(self) -> typing.Optional[str]:
        return self.get("Runtime")

    @property
    def role(self) -> typing.Optional[str]:
        return self.get("Role")

    @property
    def handler(self) -> typing.Optional[str]:
        return self.get("Handler")

    @property
    def size(self) -> typing.Optional[str]:
        return utils.to_human_readable_size(
            self.get("CodeSize"),
        )

    @property
    def timeout(self) -> typing.Optional[str]:
        return "{}s".format(self.get("Timeout"))

    @property
    def memory(self) -> typing.Optional[str]:
        return "{}MB".format(self.get("MemorySize"))

    @property
    def version(self) -> typing.Optional[str]:
        return self.get("Version")

    @property
    def environment(self) -> typing.Optional[str]:
        return self.get("Environment")

    @property
    def revision_id(self) -> typing.Optional[str]:
        return self.get("RevisionId")

    @property
    def layers(self) -> typing.List["LambdaLayerReference"]:
        return [LambdaLayerReference(item) for item in self.get("Layers") or []]

    @property
    def status(self) -> "Status":
        return Status(
            state=self.get("State"),
            reason=self.get("StateReason"),
            code=self.get("StateReasonCode"),
        )

    @property
    def update_status(self) -> "Status":
        return Status(
            state=self.get("LastUpdateStatus"),
            reason=self.get("LastUpdateStatusReason"),
            code=self.get("LastUpdateStatusReasonCode"),
        )

    def get_layer(
        self,
        name_or_arn: str,
    ) -> typing.Optional["LambdaLayerReference"]:
        """Fetches the layer reference if it exists for this function."""
        return next((v for v in self.layers if name_or_arn in (v.name, v.arn)), None)

    def get(self, key: str, default: typing.Any = None):
        """Fetches the value of the specified response key."""
        return (self.response or {}).get(key, default)


@dataclasses.dataclass(frozen=True)
class PublishedLayer:
    """Data structure containing information about a published layer."""

    #: Boto3 response returned by the layer publish operation.
    response: dict

    @property
    def arn(self) -> typing.Optional[str]:
        return (self.response or {}).get("LayerArn")

    @property
    def version(self) -> typing.Optional[str]:
        return (self.response or {}).get("Version")

    @property
    def versionedArn(self) -> typing.Optional[str]:
        return (self.response or {}).get("LayerVersionArn")



"""
Displays the configs loaded from the lambda.yaml file and fully
populated with defaults and dynamic values. Use this to inspect
and validate that the loaded configuration meets expectations
when parsed into the reviser shell.
"""
import argparse
import typing

import yaml

from reviser import interactivity


def get_completions(
    completer: "interactivity.ShellCompleter",
) -> typing.List[str]:
    """Shell auto-completes for this command."""
    return []


def populate_subparser(parser: argparse.ArgumentParser):
    """Populate parser for this command."""
    pass


def run(ex: "interactivity.Execution") -> "interactivity.Execution":
    """Execute a bundle operation on the selected function/layer targets."""
    print(yaml.safe_dump(ex.shell.context.configuration.serialize()))
    return ex.finalize(status="SUCCESS", message="Configuration displayed.")



"""
Special command to use in run command groups/macros to start interactive
command mode for the terminal. Useful when in scenarios where you wish
to prefix an interactive session with commonly executed commands. For
example, if you want to select certain targets with the select command
as part of starting the shell, you could create a run command
group/macro in your lambda.yaml that executes the select command
and then executes the shell command. This would updated the selection
and then with the shell command, start the shell in interactive mode.
Without specifying the shell command here, the run command group/macro
would just set a selection and then exit.
"""
import argparse
import typing

from reviser import interactivity


def get_completions(
    completer: "interactivity.ShellCompleter",
) -> typing.List[str]:
    """Shell auto-completes for this command."""
    return []


def populate_subparser(parser: argparse.ArgumentParser):
    """Populate parser for this command."""
    pass


def run(ex: "interactivity.Execution") -> "interactivity.Execution":
    """Exits the shell."""
    ex.shell.is_interactive = True
    return ex.finalize(
        status="SHELL",
        message="Switching to interactive shell execution.",
        echo=True,
    )



import dataclasses
import typing

from reviser.definitions import abstracts
from reviser.definitions import enumerations
from .bundling import Bundle
from .depending import Dependency
from .depending import PipDependency
from .depending import PipperDependency
from .depending import PoetryDependency
from .enviromentals import EnvironmentVariable
from .layering import AttachedLayer
from .targeting import Target


@dataclasses.dataclass(frozen=True)
class Configuration(abstracts.Specification):
    """Configuration Settings data structure."""

    @property
    def targets(self) -> typing.List["Target"]:
        """Build and deploy targets."""
        return [
            Target(
                directory=self.directory,
                data=item,
                connection=self.connection,
                configuration=self,
            )
            for item in self.get("targets", default=[])
        ]

    @property
    def function_targets(self) -> typing.List["Target"]:
        """Build and deploy function targets."""
        return [t for t in self.targets if t.kind == enumerations.TargetType.FUNCTION]

    @property
    def layer_targets(self) -> typing.List["Target"]:
        """Build and deploy layer targets."""
        return [t for t in self.targets if t.kind == enumerations.TargetType.LAYER]

    @property
    def bucket(self) -> typing.Optional[str]:
        """Retrieves the bucket to use for uploading to S3."""
        buckets = self.get(
            "buckets",
            default=self.get("bucket", default=None),
        )

        if not buckets:
            return None

        if isinstance(buckets, str):
            return buckets

        return buckets[self.connection.aws_account_id]

    @property
    def aws_region(self) -> str:
        """Name of the region that this configuration is targeting."""
        return self.get("region") or self.connection.session.region_name or "us-east-1"

    def get_function(self, name: str) -> typing.Optional["Target"]:
        """Retrieves the target with the matching name."""
        return next((t for t in self.function_targets if name in t.names), None)

    def get_layer(self, name: str) -> typing.Optional["Target"]:
        """Retrieves the target with the matching name."""
        return next((t for t in self.layer_targets if name in t.names), None)

    def serialize(self) -> dict:
        """Serializes the object for output representation."""
        return {
            "bucket": self.bucket,
            "run": self.get("run"),
            "region": self.aws_region,
            "function_targets": [t.serialize() for t in self.function_targets],
            "layer_targets": [t.serialize() for t in self.layer_targets],
        }



import os
import pathlib
import subprocess
import sys
import typing

from reviser import parsing


def main(arguments: typing.List[str] = None):
    """Start the shell container based on the command line arguments."""
    print(f"\n[DIRECTORY]: {os.path.realpath(os.curdir)}")
    args = parsing.create_parser(False).parse_args(args=arguments)

    credentials_directory = pathlib.Path(args.aws_directory).absolute()
    deploy_directory = pathlib.Path(args.root_directory).absolute()
    folder_name = deploy_directory.name
    version = args.image_tag_version

    command = [
        "docker",
        "run",
        "-it",
        "--rm",
        "-v",
        "{}:/root/.aws".format(str(credentials_directory).replace("\\", "/")),
        "-v",
        "{}:/project/{}".format(str(deploy_directory), folder_name),
        "--workdir=/project/{}".format(folder_name),
        f"swernst/reviser:{version}-{args.runtime}",
        *[a.replace("\\", "/") for a in (arguments or sys.argv)[1:]],
    ]

    display = " ".join(command).replace(" -", "\n        -")
    print(f"[RUN]: {display}\n")

    try:
        subprocess.check_call(command)
        return True
    except subprocess.CalledProcessError:
        return False


if __name__ == "__main__":
    main()



import typing


def to_human_readable_size(
    number_of_bytes: typing.Optional[int],
) -> typing.Optional[str]:
    """Converts bytes value to human readable size."""
    if number_of_bytes is None:
        return None

    step_to_greater_unit = 1024
    unit = "bytes"

    if (number_of_bytes / step_to_greater_unit) >= 1:
        number_of_bytes /= step_to_greater_unit
        unit = "KB"

    if (number_of_bytes / step_to_greater_unit) >= 1:
        number_of_bytes /= step_to_greater_unit
        unit = "MB"

    if (number_of_bytes / step_to_greater_unit) >= 1:
        number_of_bytes /= step_to_greater_unit
        unit = "GB"

    if (number_of_bytes / step_to_greater_unit) >= 1:
        number_of_bytes /= step_to_greater_unit
        unit = "TB"

    number_of_bytes = round(number_of_bytes, 1)
    return f"{number_of_bytes} {unit}"



import re
import shlex
import typing

from prompt_toolkit import completion
from prompt_toolkit.document import Document

from reviser import commands
from reviser.interactivity import shells


class ShellCompleter(completion.Completer):
    """
    Implements a concrete Prompt Toolkit Completer class specific to the
    behaviors of the lambda deployer interactive shell.
    """

    def __init__(self, shell: "shells.Shell"):
        """Initializes the prompt toolkit custom auto-completer."""
        super(ShellCompleter, self).__init__()
        self.shell = shell
        self.document: typing.Optional[Document] = None
        self.event: typing.Optional[completion.CompleteEvent] = None
        self.args: typing.List[str] = []

    @property
    def action(self) -> str:
        """Currently specified root shell command action."""
        return self.args[0] if self.args else ""

    @property
    def last(self) -> str:
        """The last argument in the current line."""
        return self.args[-1] if self.args else ""

    def get_completions(
        self, document: Document, complete_event: completion.CompleteEvent
    ) -> typing.Generator[completion.Completion, None, None]:
        """
        Implements the completion process where the Prompt Toolkit document
        and complete event are used to determine the applicable completions
        for the current shell input state.
        """
        try:
            self.args = shlex.split(document.text.strip())
        except ValueError:
            return

        self.document = document
        self.event = complete_event

        for c in _get_completions(self):
            yield c


def _get_completions(
    completer: ShellCompleter,
) -> typing.List[completion.Completion]:
    """
    Creates a list of completions that should be returned by the
    ShellCompleter.get_completions function.
    """
    if completer.document.text.lstrip().find(" ") == -1:
        return _get_action_completions(completer)

    return _get_command_completions(completer) + _get_state_completions(completer)


def _get_action_completions(completer: ShellCompleter):
    """Returns matching command actions for the current shell prompt state."""
    actions = list(commands.ALIASES.keys()) + list(commands.COMMANDS.keys())
    index = len(completer.action)
    matches = [a for a in actions if a.startswith(completer.action)]
    return [completion.Completion(m, start_position=-index) for m in matches]


def _get_command_completions(completer: ShellCompleter):
    """
    Returns completions specific to the currently specified command
    and/or sub-command in the shell. These completions are defined in the
    completer.configs module in the configs.COMMAND_COMPLETES dictionary.
    """
    command_module = commands.get_module(completer.action)
    get_completions = getattr(command_module, "get_completions", lambda x: [])
    options = get_completions(completer)
    index = len(completer.last)
    return [
        completion.Completion(item, start_position=-index)
        for item in options
        if item.startswith(completer.last)
    ]


def _get_state_completions(completer: ShellCompleter):
    """
    Returns completions based on the current state of the shell. At this
    point, this only includes lambda function/layer name completion.
    """
    index = len(completer.last)
    targets = [
        *completer.shell.context.configuration.function_targets,
        *completer.shell.context.configuration.layer_targets,
    ]
    names = [n for t in targets for n in t.names]

    pattern = re.compile(r"[\s_]+")
    names += list(
        {
            word
            for name in names
            for word in pattern.sub("-", name).split("-")
            if len(word) > 2
        }
    )

    return [
        completion.Completion(n, start_position=-index)
        for n in names
        if n.startswith(completer.last)
    ]



import dataclasses
import typing


@dataclasses.dataclass(frozen=True)
class Selection:
    """
    Data structure that defines a selection of targets on which
    to act within the loaded context.
    """

    bundle_all: bool = True
    function_needles: typing.List[str] = dataclasses.field(default_factory=lambda: [])
    function_names: typing.List[str] = dataclasses.field(default_factory=lambda: [])
    layer_needles: typing.List[str] = dataclasses.field(default_factory=lambda: [])
    layer_names: typing.List[str] = dataclasses.field(default_factory=lambda: [])
